#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include "Socket.H"
#include "Logger.H"

extern Logger logger;

Socket::Socket(const string &host, unsigned short port, long flags):
		_sockfd(-1),
		_errno(0),
		_flags(flags),
		_outputstream(NULL),
		_inputstream(NULL),
		_host(host),
		_port(port) {
	memset(&_serv_addr, sizeof(_serv_addr), 0);
	pthread_mutex_init(&_rlock,NULL);
	pthread_mutex_init(&_wlock,NULL);
}

Socket::~Socket() {
	close();
	pthread_mutex_destroy(&_rlock);
	pthread_mutex_destroy(&_wlock);
}

int Socket::open() {
	struct hostent
		*hp;

	pthread_mutex_lock(&_rlock);
	if ((_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		_errno=errno;
		return -1;
	}
	if ((hp = gethostbyname(_host.c_str())) == NULL) {
		_errno=errno;
		::close(_sockfd);
		return -2;
	}
	_serv_addr.sin_port = htons(_port);
	_serv_addr.sin_family = AF_INET;
	memcpy(&(_serv_addr.sin_addr.s_addr), hp->h_addr, hp->h_length);
	if ((connect(_sockfd,(sockaddr *)&_serv_addr, sizeof(_serv_addr))) < 0) {
		_errno=errno;
		::close(_sockfd);
		return -3;
	}
	activateFlags();
	_outputstream = fdopen(_sockfd,"w");
	_inputstream = fdopen(_sockfd,"r");
	setbuf(_outputstream,NULL);
	setbuf(_inputstream,NULL);
	pthread_mutex_unlock(&_rlock);
	return 0;
}

void Socket::close() {
	pthread_mutex_lock(&_wlock);
	_outputstream = NULL;
	_inputstream = NULL;
	if (_sockfd != -1) {
		::close(_sockfd);
		_sockfd=-1;
	}
	_sockfd = -1;
	pthread_mutex_unlock(&_wlock);
}

int Socket::read(char *buf, unsigned int count, int waitsecs) {
	unsigned int
		rcvd=0;
	int
		secs=0,
		rv;

	pthread_mutex_lock(&_rlock);
	while (rcvd < count) {
		if ((rv=::read(_sockfd, buf, (count-rcvd))) < 0) {
			_errno=errno;
			if (_errno == EINTR) {
				continue;
			}
			else if (_errno == EAGAIN) {
				if (_flags & O_NONBLOCK) {
					if ((waitsecs > secs++) || (waitsecs==-1)) {
						usleep(50000);
						continue;
					}
				}
				break;
			}
			else {
				logger.log("Socket::read: got wierd error: %s",strerror(_errno));
				return -1;
			}
		}
		else if (rv == 0) {
			break;
		}
		rcvd+=rv;
		buf+=rv;
	}
	pthread_mutex_unlock(&_rlock);
	return (rcvd);
}

int Socket::write(const char *buf, unsigned int len, int waitsecs) {
	unsigned int
		sent=0;
	int
		secs=0,
		rv;

	pthread_mutex_lock(&_wlock);
	while (sent < len) {
		if ((rv=::write(_sockfd, buf, (len-sent))) < 0) {
			_errno=errno;
			if (_errno == EINTR) {
				logger.log("Socket::write: got EINTR");
				continue;
			}
			else if (_errno == EAGAIN) {
				logger.log("Socket::write: got EAGAIN");
				if ((waitsecs=-1) && (_flags & O_NONBLOCK))
					continue;
				if (waitsecs > secs++) {
					sleep(1);
					continue;
				}
				break;
			}
			else {
				return -1;
			}
		}
		else if (rv == 0) {
			logger.log("Socket::write: got EOF");
			break;
		}
		sent+=rv;
		buf+=rv;
	}
	if (sent!=len) {
		logger.log("Socket::write: sent=%u len=%u\n",sent,len);
	}
	pthread_mutex_unlock(&_wlock);
	return (sent);
}

int Socket::setNonBlock() {
	_flags |= O_NONBLOCK;
	if (_sockfd != -1) {
		activateFlags();
	}
	return 0;
}

int Socket::activateFlags() {
	long flags = fcntl(_sockfd,F_GETFL);
	fcntl(_sockfd,F_SETFL,flags | _flags);
	return 0;
}
